import io
import pathlib
import typing
import unittest
from collections import defaultdict
from contextlib import redirect_stdout
from unittest.mock import MagicMock, Mock, patch

import pytest

from runem.config_metadata import ConfigMetadata
from runem.config_parse import (
    _parse_global_config,
    _parse_job,
    load_config_metadata,
    parse_hook_config,
    parse_job_config,
)
from runem.types.common import JobNames, JobPhases, JobTags, OrderedPhases
from runem.types.errors import FunctionNotFound, SystemExitBad
from runem.types.hooks import HookName
from runem.types.runem_config import (
    Config,
    GlobalConfig,
    GlobalSerialisedConfig,
    HookConfig,
    HookSerialisedConfig,
    JobConfig,
    JobSerialisedConfig,
    PhaseGroupedJobs,
)


def test_parse_job_config() -> None:
    """Tests basic parsing of the job config."""
    job_config: JobConfig = {
        "addr": {
            "file": __file__,
            "function": "test_parse_job_config",
        },
        "label": "reformat py",
        "when": {
            "phase": "edit",
            "tags": set(
                (
                    "py",
                    "format",
                )
            ),
        },
    }
    tags: JobTags = set(["py"])
    jobs_by_phase: PhaseGroupedJobs = defaultdict(list)
    job_names: JobNames = set()
    phases: JobPhases = set()
    phase_order: OrderedPhases = ()
    parse_job_config(
        cfg_filepath=pathlib.Path(__file__),
        job=job_config,
        in_out_tags=tags,
        in_out_jobs_by_phase=jobs_by_phase,
        in_out_job_names=job_names,
        in_out_phases=phases,
        phase_order=phase_order,
    )
    assert tags == {"format", "py"}
    assert jobs_by_phase == {
        "edit": [
            {
                "addr": {
                    "file": "test_config_parse.py",
                    "function": "test_parse_job_config",
                },
                "label": "reformat py",
                "when": {"phase": "edit", "tags": set(("py", "format"))},
            }
        ]
    }
    assert job_names == {"reformat py"}
    assert phases == {"edit"}


def test_parse_job_config_handles_multiple_cwd() -> None:
    """Tests that multiple cwd generate jobs per cwd."""
    job_config: JobConfig = {
        "addr": {
            "file": __file__,
            "function": "test_parse_job_config",
        },
        "ctx": {"cwd": ["path/a", "path/b"]},
        "label": "reformat py",
        "when": {
            "phase": "edit",
            "tags": set(
                (
                    "py",
                    "format",
                )
            ),
        },
    }
    tags: JobTags = set(["py"])
    jobs_by_phase: PhaseGroupedJobs = defaultdict(list)
    job_names: JobNames = set()
    phases: JobPhases = set()
    phase_order: OrderedPhases = ()
    parse_job_config(
        cfg_filepath=pathlib.Path(__file__),
        job=job_config,
        in_out_tags=tags,
        in_out_jobs_by_phase=jobs_by_phase,
        in_out_job_names=job_names,
        in_out_phases=phases,
        phase_order=phase_order,
    )
    assert tags == {"a", "b", "format", "py"}, "tags should include the explicit"
    assert jobs_by_phase == {
        "edit": [
            {
                "addr": {
                    "file": "test_config_parse.py",
                    "function": "test_parse_job_config",
                },
                "ctx": {"cwd": "path/a"},
                "label": "reformat py(path/a)",
                "when": {"phase": "edit", "tags": set(("a", "py", "format"))},
            },
            {
                "addr": {
                    "file": "test_config_parse.py",
                    "function": "test_parse_job_config",
                },
                "ctx": {"cwd": "path/b"},
                "label": "reformat py(path/b)",
                "when": {"phase": "edit", "tags": set(("b", "py", "format"))},
            },
        ]
    }
    assert job_names == {"reformat py(path/a)", "reformat py(path/b)"}
    assert phases == {"edit"}


def test_parse_job_config_throws_on_dupe_name() -> None:
    """Tests for job-name clashes."""
    job_config: JobConfig = {
        "addr": {
            "file": __file__,
            "function": "test_parse_job_config",
        },
        "label": "reformat py",
        "when": {
            "phase": "edit",
            "tags": set(
                (
                    "py",
                    "format",
                )
            ),
        },
    }
    tags: JobTags = set(["py"])
    jobs_by_phase: PhaseGroupedJobs = defaultdict(list)
    job_names: JobNames = set()
    phases: JobPhases = set()
    phase_order: OrderedPhases = ()

    # first call should be fine
    parse_job_config(
        cfg_filepath=pathlib.Path(__file__),
        job=job_config,
        in_out_tags=tags,
        in_out_jobs_by_phase=jobs_by_phase,
        in_out_job_names=job_names,
        in_out_phases=phases,
        phase_order=phase_order,
    )
    assert job_config["label"] in job_names

    # second call should error
    with pytest.raises(SystemExit):
        parse_job_config(
            cfg_filepath=pathlib.Path(__file__),
            job=job_config,
            in_out_tags=tags,
            in_out_jobs_by_phase=jobs_by_phase,
            in_out_job_names=job_names,
            in_out_phases=phases,
            phase_order=phase_order,
        )


def test_parse_job_config_throws_on_missing_key() -> None:
    """Tests for expected keys are reported if missing."""
    job_config: JobConfig = {
        "addr": {  # type: ignore[typeddict-item]
            "file": __file__,
            # intentionally removed:
            # "function": "test_parse_job_config",
        },
        "label": "reformat py",
        "when": {
            "phase": "edit",
            "tags": set(
                (
                    "py",
                    "format",
                )
            ),
        },
    }
    tags: JobTags = set(["py"])
    jobs_by_phase: PhaseGroupedJobs = defaultdict(list)
    job_names: JobNames = set()
    phases: JobPhases = set()
    phase_order: OrderedPhases = ()
    with pytest.raises(ValueError) as err_info:
        parse_job_config(
            cfg_filepath=pathlib.Path(__file__),
            job=job_config,
            in_out_tags=tags,
            in_out_jobs_by_phase=jobs_by_phase,
            in_out_job_names=job_names,
            in_out_phases=phases,
            phase_order=phase_order,
        )
    assert str(err_info.value).startswith(
        ("job config entry is missing 'function' data")
    )


def test_parse_global_config_empty() -> None:
    """Test the global config parse handles empty data."""
    dummy_global_config: GlobalConfig = {
        "phases": tuple(),
        "min_version": None,
        "options": [],
        "files": [],
    }
    phases, options, file_filters = _parse_global_config(dummy_global_config)
    assert phases == tuple()
    assert options == tuple()
    assert not file_filters


def test_parse_global_config_missing() -> None:
    """Test the global config parse handles missing data."""
    dummy_global_config: GlobalConfig = {  # type: ignore
        "phases": tuple(),
        # intentionally missing: "options": [],
        # intentionally missing: "files": [],
    }
    phases, options, file_filters = _parse_global_config(dummy_global_config)
    assert phases == tuple()
    assert options == tuple()
    assert not file_filters


def test_parse_global_config_full() -> None:
    """Test the global config parse handles missing data."""
    dummy_global_config: GlobalConfig = {
        "phases": tuple(),
        "min_version": None,
        "options": [
            {
                "option": {
                    "name": "dummy option",
                    "aliases": None,
                    "default": False,
                    "type": "bool",
                    "desc": "dummy description",
                }
            }
        ],
        "files": [{"filter": {"tag": "dummy tag", "regex": ".*"}}],
    }
    phases, options, file_filters = _parse_global_config(dummy_global_config)
    assert phases == tuple()
    assert options == (
        {
            "name": "dummy option",
            "aliases": None,
            "default": False,
            "type": "bool",
            "desc": "dummy description",
        },
    )
    assert file_filters == {"dummy tag": {"regex": ".*", "tag": "dummy tag"}}


@pytest.mark.parametrize(
    "do_hooks",
    [
        True,
        False,
    ],
)
def test_load_config_metadata(do_hooks: bool) -> None:
    """Test parsing works for a full config."""
    global_config: GlobalSerialisedConfig = {
        "config": {
            "phases": ("dummy phase 1",),
            "files": [],
            "min_version": None,
            "options": [],
        }
    }
    job_config: JobSerialisedConfig = {
        "job": {
            "addr": {
                "file": __file__,
                "function": "test_load_config_metadata",
            },
            "label": "dummy job label",
            "when": {
                "phase": "dummy phase 1",
                "tags": set(
                    (
                        "dummy tag 1",
                        "dummy tag 2",
                    )
                ),
            },
        }
    }
    full_config: Config = [global_config, job_config]
    if do_hooks:
        # optionally do hooks, to capture verbose logging without having to
        # write another test.
        hook_config: HookSerialisedConfig = {
            "hook": {
                "command": "echo 'test hook command'",
                "hook_name": HookName("on-exit"),
            }
        }
        full_config.append(hook_config)

    config_file_path = pathlib.Path(__file__).parent / ".runem.yml"
    expected_job: JobConfig = {
        "addr": {
            "file": "test_config_parse.py",
            "function": "test_load_config_metadata",
        },
        "label": "dummy job label",
        "when": {
            "phase": "dummy phase 1",
            "tags": {"dummy tag 1", "dummy tag 2"},
        },
    }
    expected_jobs: PhaseGroupedJobs = defaultdict(list)
    expected_jobs["dummy phase 1"] = [
        expected_job,
    ]
    expected_config_metadata: ConfigMetadata = ConfigMetadata(
        cfg_filepath=config_file_path,
        phases=("dummy phase 1",),
        options_config=tuple(),
        file_filters={
            # "dummy tag": {
            #     "tag": "dummy tag",
            #     "regex": ".*1.txt",  # should match just one file
            # }
        },
        hook_manager=MagicMock(),
        jobs=expected_jobs,
        all_job_names=set(("dummy job label",)),
        all_job_phases=set(("dummy phase 1",)),
        all_job_tags=set(
            (
                "dummy tag 2",
                "dummy tag 1",
            )
        ),
    )

    with io.StringIO() as buf, redirect_stdout(buf):
        result: ConfigMetadata = load_config_metadata(full_config, config_file_path, [])
        stdout = buf.getvalue().split("\n")
    assert stdout == [""]
    assert result.phases == expected_config_metadata.phases
    assert result.options_config == expected_config_metadata.options_config
    assert result.file_filters == expected_config_metadata.file_filters
    assert result.jobs == expected_config_metadata.jobs
    assert result.all_job_names == expected_config_metadata.all_job_names
    assert result.all_job_phases == expected_config_metadata.all_job_phases
    assert result.all_job_tags == expected_config_metadata.all_job_tags


def test_load_config_metadata_hooks() -> None:
    """Test parsing works for a full config, with user hook."""
    global_config: GlobalSerialisedConfig = {
        "config": {
            "phases": ("dummy phase 1",),
            "files": [],
            "min_version": None,
            "options": [],
        }
    }
    job_config: JobSerialisedConfig = {
        "job": {
            "addr": {
                "file": __file__,
                "function": "test_load_config_metadata",
            },
            "label": "dummy job label",
            "when": {
                "phase": "dummy phase 1",
                "tags": set(
                    (
                        "dummy tag 1",
                        "dummy tag 2",
                    )
                ),
            },
        }
    }
    hook_config: HookSerialisedConfig = {
        "hook": {
            "command": "echo 'test hook command'",
            "hook_name": HookName("on-exit"),
        }
    }
    user_hook_config_1: HookSerialisedConfig = {
        "hook": {
            "command": "echo 'test user-hook command 1'",
            "hook_name": HookName("on-exit"),
        }
    }
    user_hook_config_2: HookSerialisedConfig = {
        "hook": {
            "command": "echo 'test user-hook command 2'",
            "hook_name": HookName("on-exit"),
        }
    }
    user_jobs_only_config: JobSerialisedConfig = {
        "job": {
            "command": "echo 'should not be seen'",
            "label": "SHOULD NOT SEE ME",
        }
    }
    full_config: Config = [global_config, job_config, hook_config]
    config_file_path = pathlib.Path(__file__).parent / ".runem.yml"
    expected_job: JobConfig = {
        "addr": {
            "file": "test_config_parse.py",
            "function": "test_load_config_metadata",
        },
        "label": "dummy job label",
        "when": {
            "phase": "dummy phase 1",
            "tags": {"dummy tag 1", "dummy tag 2"},
        },
    }
    expected_jobs: PhaseGroupedJobs = defaultdict(list)
    expected_jobs["dummy phase 1"] = [
        expected_job,
    ]
    expected_config_metadata: ConfigMetadata = ConfigMetadata(
        cfg_filepath=config_file_path,
        phases=("dummy phase 1",),
        options_config=tuple(),
        file_filters={
            # "dummy tag": {
            #     "tag": "dummy tag",
            #     "regex": ".*1.txt",  # should match just one file
            # }
        },
        hook_manager=MagicMock(),
        jobs=expected_jobs,
        all_job_names=set(("dummy job label",)),
        all_job_phases=set(("dummy phase 1",)),
        all_job_tags=set(
            (
                "dummy tag 2",
                "dummy tag 1",
            )
        ),
    )

    user_config_1: Config = [
        user_hook_config_1,
    ]
    user_config_2: Config = [
        user_hook_config_2,
    ]
    user_config_jobs_only: Config = [
        user_jobs_only_config,
    ]

    result: ConfigMetadata = load_config_metadata(
        full_config,
        config_file_path,
        [
            (user_config_1, pathlib.Path(__file__)),
            (user_config_2, pathlib.Path(__file__)),
            (user_config_jobs_only, pathlib.Path(__file__)),
        ],
    )
    assert result.phases == expected_config_metadata.phases
    assert result.options_config == expected_config_metadata.options_config
    assert result.file_filters == expected_config_metadata.file_filters
    assert result.jobs == expected_config_metadata.jobs
    assert result.all_job_names == expected_config_metadata.all_job_names
    assert result.all_job_phases == expected_config_metadata.all_job_phases
    assert result.all_job_tags == expected_config_metadata.all_job_tags


def test_load_config_metadata_raises_on_invalid() -> None:
    """Test throws for an invalid config."""
    invalid_config_spec: GlobalSerialisedConfig = {  # type: ignore
        "invalid": None,
    }
    invalid_config: Config = [
        invalid_config_spec,
    ]
    config_file_path = pathlib.Path(__file__).parent / ".runem.yml"

    with pytest.raises(RuntimeError):
        load_config_metadata(invalid_config, config_file_path, [])


def test_load_config_metadata_duplicated_global_raises() -> None:
    """Test the global config parse raises with duplicated global config."""
    dummy_global_config: GlobalSerialisedConfig = {
        "config": {
            "phases": ("dummy phase 1",),
            "min_version": None,
            "options": [
                {
                    "option": {
                        "name": "dummy option",
                        "aliases": None,
                        "default": False,
                        "type": "bool",
                        "desc": "dummy description",
                    }
                }
            ],
            "files": [{"filter": {"tag": "dummy tag", "regex": ".*"}}],
        }
    }
    invalid_config: Config = [
        dummy_global_config,
        dummy_global_config,
    ]
    config_file_path = pathlib.Path(__file__).parent / ".runem.yml"
    with pytest.raises(ValueError):
        load_config_metadata(invalid_config, config_file_path, [])


def test_load_config_metadata_empty_phases_raises() -> None:
    """Test the global config raises if the phases are empty."""
    dummy_global_config: GlobalSerialisedConfig = {
        "config": {
            "phases": (),
            "min_version": None,
            "options": [
                {
                    "option": {
                        "name": "dummy option",
                        "aliases": None,
                        "default": False,
                        "type": "bool",
                        "desc": "dummy description",
                    }
                }
            ],
            "files": [{"filter": {"tag": "dummy tag", "regex": ".*"}}],
        }
    }
    invalid_config: Config = [
        dummy_global_config,
        dummy_global_config,
    ]
    config_file_path = pathlib.Path(__file__).parent / ".runem.yml"
    with pytest.raises(ValueError):
        load_config_metadata(invalid_config, config_file_path, [])


def test_load_config_metadata_missing_phases_raises() -> None:
    """Test the global config raises if the phases are missing."""
    dummy_global_config: GlobalSerialisedConfig = {
        "config": {  # type: ignore
            "options": [
                {
                    "option": {
                        "name": "dummy option",
                        "aliases": None,
                        "default": False,
                        "type": "bool",
                        "desc": "dummy description",
                    }
                }
            ],
            "files": [{"filter": {"tag": "dummy tag", "regex": ".*"}}],
        }
    }
    invalid_config: Config = [
        dummy_global_config,
        dummy_global_config,
    ]
    config_file_path = pathlib.Path(__file__).parent / ".runem.yml"
    with pytest.raises(ValueError):
        load_config_metadata(invalid_config, config_file_path, [])


@patch(
    "runem.config_parse._parse_global_config",
    return_value=(None, (), {}),
)
def test_load_config_metadata_warning_if_missing_phase_order(
    mock_parse_global_config: unittest.mock.Mock,
) -> None:
    """Test the global config raises if the phases are missing."""
    dummy_global_config: GlobalSerialisedConfig = {
        "config": {  # type: ignore
            "options": [
                {
                    "option": {
                        "name": "dummy option",
                        "aliases": None,
                        "default": False,
                        "type": "bool",
                        "desc": "dummy description",
                    }
                }
            ],
            "files": [{"filter": {"tag": "dummy tag", "regex": ".*"}}],
        }
    }
    valid_config: Config = [
        dummy_global_config,
    ]
    config_file_path = pathlib.Path(__file__).parent / ".runem.yml"

    # run the command and capture output
    with io.StringIO() as buf, redirect_stdout(buf):
        load_config_metadata(valid_config, config_file_path, [])
        run_command_stdout = buf.getvalue()

    assert run_command_stdout.split("\n") == [
        "runem: WARNING: phase ordering not configured! Runs will be non-deterministic!",
        "",
    ]
    mock_parse_global_config.assert_called()


@patch(
    "runem.config_parse.get_job_wrapper",
    return_value=None,
)
def test_parse_job_with_tags(mock_get_job_wrapper: Mock) -> None:
    """Test case where job_tags is not empty."""
    cfg_filepath = pathlib.Path(__file__)
    job_config: JobConfig = {
        "label": "Job1",
        "when": {
            "tags": {
                "tag1",
                "tag2",
            }
        },
    }
    in_out_tags: JobTags = set()
    in_out_jobs_by_phase: PhaseGroupedJobs = defaultdict(list)
    in_out_job_names: JobNames = set()
    in_out_phases: JobPhases = set()
    phase_order: OrderedPhases = ("phase1", "phase2")

    with io.StringIO() as buf, redirect_stdout(buf):
        _parse_job(
            cfg_filepath,
            job_config,
            in_out_tags,
            in_out_jobs_by_phase,
            in_out_job_names,
            in_out_phases,
            phase_order,
        )
        run_command_stdout = buf.getvalue().split("\n")

    mock_get_job_wrapper.assert_called_once()

    assert run_command_stdout == [
        "runem: WARNING: no phase found for 'Job1', using 'phase1'",
        "",
    ]
    assert in_out_tags == {"tag1", "tag2"}


@patch(
    "runem.config_parse.get_job_wrapper",
    return_value=None,
)
def test_parse_job_without_tags(mock_get_job_wrapper: Mock) -> None:
    """Test case where job_tags is empty or None."""
    cfg_filepath = pathlib.Path(__file__)
    job_config: JobConfig = {
        "label": "Job2",
        "when": {
            "phase": "phase1",
        },
    }
    in_out_tags: JobTags = set()
    in_out_jobs_by_phase: PhaseGroupedJobs = defaultdict(list)
    in_out_job_names: JobNames = set()
    in_out_phases: JobPhases = set()
    phase_order: OrderedPhases = ("phase1", "phase2")

    with io.StringIO() as buf, redirect_stdout(buf):
        _parse_job(
            cfg_filepath,
            job_config,
            in_out_tags,
            in_out_jobs_by_phase,
            in_out_job_names,
            in_out_phases,
            phase_order,
        )
        run_command_stdout = buf.getvalue().split("\n")

    mock_get_job_wrapper.assert_called_once()

    assert run_command_stdout == [""]
    assert not in_out_tags


@patch(
    "runem.config_parse.get_job_wrapper",
    side_effect=FunctionNotFound("a test message"),
)
def test_parse_job_with_bad_function(mock_get_job_wrapper: Mock) -> None:
    """Test case where job_tags is empty or None."""
    cfg_filepath = pathlib.Path(__file__)
    job_config: JobConfig = {
        "label": "Job2",
        "when": {
            "phase": "phase1",
        },
    }
    in_out_tags: JobTags = set()
    in_out_jobs_by_phase: PhaseGroupedJobs = defaultdict(list)
    in_out_job_names: JobNames = set()
    in_out_phases: JobPhases = set()
    phase_order: OrderedPhases = ("phase1", "phase2")

    with io.StringIO() as buf, redirect_stdout(buf):
        with pytest.raises(SystemExitBad):
            _parse_job(
                cfg_filepath,
                job_config,
                in_out_tags,
                in_out_jobs_by_phase,
                in_out_job_names,
                in_out_phases,
                phase_order,
            )
        run_command_stdout = buf.getvalue().split("\n")

    mock_get_job_wrapper.assert_called_once()

    assert run_command_stdout == [
        "runem: ERROR: Whilst loading job 'Job2'. a test message",
        "",
    ]
    assert not in_out_tags


@pytest.mark.parametrize(
    "phase_order",
    [
        ("phase1", "phase2"),
        (),
    ],
)
@patch(
    "runem.config_parse.get_job_wrapper",
    return_value=None,
)
def test_parse_job_with_missing_phase(
    mock_get_job_wrapper: Mock,
    phase_order: OrderedPhases,
) -> None:
    """Test case where job_tags is empty or None."""
    cfg_filepath = pathlib.Path(__file__)
    job_config: JobConfig = {
        "label": "Job2",
        "when": {
            # "phase": "phase1",
        },
    }
    in_out_tags: JobTags = set()
    in_out_jobs_by_phase: PhaseGroupedJobs = defaultdict(list)
    in_out_job_names: JobNames = set()
    in_out_phases: JobPhases = set()

    with io.StringIO() as buf, redirect_stdout(buf):
        _parse_job(
            cfg_filepath,
            job_config,
            in_out_tags,
            in_out_jobs_by_phase,
            in_out_job_names,
            in_out_phases,
            phase_order,
            warn_missing_phase=False,
        )
        run_command_stdout = buf.getvalue().split("\n")

    mock_get_job_wrapper.assert_called_once()

    assert run_command_stdout == [""]
    assert not in_out_tags


def test_parse_hook_config_with_valid_config() -> None:
    cfg_filepath = pathlib.Path(__file__)
    hook: HookConfig = {
        "command": "echo 'test hook command'",
        "hook_name": HookName("on-exit"),
    }
    # should execute without raising
    parse_hook_config(hook, cfg_filepath)


def test_parse_hook_config_with_in_valid_hook_name() -> None:
    cfg_filepath = pathlib.Path(__file__)
    hook: HookConfig = {
        "command": "echo 'test hook command'",
        "hook_name": "bad-hook-name",  # type: ignore
    }
    with pytest.raises(ValueError):
        parse_hook_config(hook, cfg_filepath)


def test_parse_hook_config_with_missing_hook_name() -> None:
    cfg_filepath = pathlib.Path(__file__)
    hook: HookConfig = {
        "command": "echo 'test hook command'",
        # "hook_name": HookName("on-exit"),
    }
    with pytest.raises(ValueError):
        parse_hook_config(hook, cfg_filepath)


def test_parse_hook_config_with_missing_command() -> None:
    cfg_filepath = pathlib.Path(__file__)
    hook: HookConfig = {
        # "command": "echo 'test hook command'",
        "hook_name": HookName("on-exit"),
    }
    with pytest.raises(ValueError):
        parse_hook_config(hook, cfg_filepath)


def test_parse_hook_config_with_bad_function_address() -> None:
    cfg_filepath = pathlib.Path(__file__)
    hook: HookConfig = {
        # "command": "echo 'test hook command'",
        "addr": {
            "file": __file__,
            "function": "non_existent_function",
        },
        "hook_name": HookName("on-exit"),
    }
    with pytest.raises(SystemExitBad):
        parse_hook_config(hook, cfg_filepath)


@pytest.mark.parametrize(
    "hook_type",
    [
        "non-existent-hook",
        123,
    ],
)
def test_parse_hook_config_with_bad_hook_name_types(hook_type: typing.Any) -> None:
    cfg_filepath = pathlib.Path(__file__)
    hook: HookConfig = {
        "command": "echo 'test hook command'",
        "hook_name": hook_type,
    }
    with pytest.raises(ValueError):
        parse_hook_config(hook, cfg_filepath)



import pathlib
import typing

from runem.types.common import JobName, JobTags, OrderedPhases, PhaseName
from runem.types.filters import TagFileFilter
from runem.types.hooks import HookName


class OptionConfig(typing.TypedDict, total=False):
    """Spec for configuring job option overrides."""

    name: str
    aliases: typing.Optional[typing.List[str]]
    alias: typing.Optional[str]
    default: bool
    type: str
    desc: typing.Optional[str]


OptionConfigs = typing.Tuple[OptionConfig, ...]


class OptionConfigSerialised(typing.TypedDict):
    """Supports better serialisation of options."""

    option: OptionConfig


class JobParamConfig(typing.TypedDict):
    """Configures what parameters are passed to the test-callable.

    FIXME: this isn't actually used at all, yet
    """

    limitFilesToGroup: bool  # whether to limit file-set for the job


class JobAddressConfig(typing.TypedDict):
    """Configuration which described a callable to call."""

    file: str  # the file-module where 'function' can be found
    function: str  # the 'function' in module to run


class JobContextConfig(typing.TypedDict, total=False):
    # what parameters the job needs # DEFUNCT
    params: typing.Optional[JobParamConfig]

    # the path or paths to run the command in. If given a list the job will be
    # duplicated for each given path.
    cwd: typing.Optional[typing.Union[str, typing.List[str]]]


class JobWhen(typing.TypedDict, total=False):
    """Configures WHEN to call the callable i.e. priority."""

    tags: JobTags  # the job tags - used for filtering job-types
    phase: PhaseName  # the phase when the job should be run


class JobWrapper(typing.TypedDict, total=False):
    """A base-type for jobs, hooks, and things that can be invoked."""

    addr: JobAddressConfig  # which callable to call
    command: str  # a one-liner command to be run


class JobConfig(JobWrapper, total=False):
    """A dict that defines a job to be run.

    It consists of the label, address, context and filter information

    TODO: make a class variant of this
    """

    label: JobName  # the name of the job
    ctx: typing.Optional[JobContextConfig]  # how to call the callable
    when: JobWhen  # when to call the job


Jobs = typing.List[JobConfig]


class TagFileFilterSerialised(typing.TypedDict):
    """Supports better serialisation of TagFileFilters."""

    filter: TagFileFilter


class GlobalConfig(typing.TypedDict):
    """The config for the entire test run."""

    # Phases control the order of jobs, jobs earlier in the stack get run earlier
    # the core ide here is to ensure that certain types of job-dependencies,
    # such as code-reformatting jobs run before analysis tools, therefore making
    # any error messages about the code give consistent line numbers e..g if a
    # re-formatter edits a file the error line will move and the analysis phase
    # will report the wrong line.
    phases: OrderedPhases

    # Options control the extra flags that are optionally consumed by job.
    # Options configured here are used to set command-line-options. All options
    # and their current state are passed to each job.
    options: typing.Optional[typing.List[OptionConfigSerialised]]

    # File filters control which files will be passed to jobs for a given tags.
    # Job will receive the super-set of files for all that job's tags.
    files: typing.Optional[typing.List[TagFileFilterSerialised]]

    # Which minimal version of runem does this config support?
    min_version: typing.Optional[str]


class GlobalSerialisedConfig(typing.TypedDict):
    """Intended to make reading a config file easier.

    Unlike JobSerialisedConfig, this type may not actually help readability.

    An intermediary type for serialisation of the global config, the 'global' resides
    inside a 'global' key and therefore is easier to find and reason about.
    """

    config: GlobalConfig


class HookConfig(JobWrapper, total=False):
    """Specification for hooks.

    Like JobConfig with use addr or command to specify what to execute.
    """

    hook_name: HookName  # the hook for when this is called


class HookSerialisedConfig(typing.TypedDict):
    """Intended to make reading a config file easier.

    Also, unlike JobSerialisedConfig, this type may not actually help readability.
    """

    hook: HookConfig


class JobSerialisedConfig(typing.TypedDict):
    """Makes serialised configs easier to read.

    An intermediary typ for serialisation as each 'job' resides inside a 'job' key.

    This makes formatting of YAML config _significantly_ easier to understand.
    """

    job: JobConfig


ConfigNodes = typing.Union[
    GlobalSerialisedConfig, JobSerialisedConfig, HookSerialisedConfig
]
# The config format as it is serialised to/from disk
Config = typing.List[ConfigNodes]
UserConfigMetadata = typing.List[typing.Tuple[Config, pathlib.Path]]
Hooks = typing.DefaultDict[HookName, typing.List[HookConfig]]
# A dictionary to hold hooks, with hook names as keys
HooksStore = typing.Dict[HookName, typing.List[HookConfig]]
PhaseGroupedJobs = typing.DefaultDict[PhaseName, Jobs]



import os
import pathlib
import typing
import uuid
from datetime import timedelta
from timeit import default_timer as timer

from typing_extensions import Unpack

from runem.config_metadata import ConfigMetadata
from runem.informative_dict import ReadOnlyInformativeDict
from runem.job import Job
from runem.job_wrapper import get_job_wrapper
from runem.log import error, log, warn
from runem.types.common import FilePathList, JobTags
from runem.types.filters import FilePathListLookup
from runem.types.runem_config import JobConfig
from runem.types.types_jobs import (
    AllKwargs,
    HookSpecificKwargs,
    JobFunction,
    JobKwargs,
    JobReturn,
    JobTiming,
    TimingEntries,
    TimingEntry,
)


def job_execute_inner(
    job_config: JobConfig,
    config_metadata: ConfigMetadata,
    file_lists: FilePathListLookup,
    **kwargs: Unpack[HookSpecificKwargs],
) -> typing.Tuple[JobTiming, JobReturn]:
    """Wrapper for running a job inside a sub-process.

    Returns the time information and any reports the job generated
    """
    label = Job.get_job_name(job_config)
    if config_metadata.args.verbose:
        log(f"START: '{label}'")
    root_path: pathlib.Path = config_metadata.cfg_filepath.parent
    job_tags: typing.Optional[JobTags] = Job.get_job_tags(job_config)
    os.chdir(root_path)
    function: JobFunction = get_job_wrapper(job_config, config_metadata.cfg_filepath)

    # get the files for all files found for this job's tags
    file_list: FilePathList = Job.get_job_files(file_lists, job_tags)

    if not file_list:
        # no files to work on
        if not config_metadata.args.silent:
            warn(f"skipping job '{label}', no files for job")
        return {
            "job": (f"{label}: no files!", timedelta(0)),
            "commands": [],
        }, None

    sub_command_timings: TimingEntries = []

    def _record_sub_job_time(label: str, timing: timedelta) -> None:
        """Record timing information for sub-commands/tasks, atomically.

        For example inside of run_command() calls
        """
        sub_command_timings.append((label, timing))

    if (
        "ctx" in job_config
        and job_config["ctx"] is not None
        and "cwd" in job_config["ctx"]
        and job_config["ctx"]["cwd"]
    ):
        assert isinstance(job_config["ctx"]["cwd"], str)
        os.chdir(root_path / job_config["ctx"]["cwd"])
    else:
        os.chdir(root_path)

    start = timer()
    if config_metadata.args.verbose:
        log(f"job: running: '{Job.get_job_name(job_config)}'")
    reports: JobReturn
    try:
        # Define the common args for all jobs and hooks.
        job_k_args: JobKwargs = {
            "config_metadata": config_metadata,
            "file_list": file_list,
            "job": job_config,
            "label": Job.get_job_name(job_config),
            "options": ReadOnlyInformativeDict(config_metadata.options),
            "procs": config_metadata.args.procs,
            "record_sub_job_time": _record_sub_job_time,
            "root_path": root_path,
            "verbose": config_metadata.args.verbose,
        }
        # Merge in the hook-specific kwargs (if any) for situations where we are
        # calling hooks.
        all_k_args: AllKwargs = {
            **job_k_args,
            **kwargs,
        }

        assert isinstance(function, JobFunction)
        reports = function(**all_k_args)
    except BaseException:  # pylint: disable=broad-exception-caught
        # log that we hit an error on this job and re-raise
        log(prefix=False)
        error(f"job: job '{Job.get_job_name(job_config)}' failed to complete!")
        # re-raise
        raise

    end = timer()
    time_taken: timedelta = timedelta(seconds=end - start)
    if config_metadata.args.verbose:
        log(f"job: DONE: '{label}': {time_taken}")
    this_job_timing_data: TimingEntry = (label, time_taken)
    return ({"job": this_job_timing_data, "commands": sub_command_timings}, reports)


def job_execute(
    job_config: JobConfig,
    running_jobs: typing.Dict[str, str],
    completed_jobs: typing.Dict[str, str],
    config_metadata: ConfigMetadata,
    file_lists: FilePathListLookup,
    **kwargs: Unpack[HookSpecificKwargs],
) -> typing.Tuple[JobTiming, JobReturn]:
    """Thin-wrapper around job_execute_inner needed for mocking in tests.

    Needed for faster tests.
    """
    this_id: str = str(uuid.uuid4())
    running_jobs[this_id] = Job.get_job_name(job_config)
    try:
        results = job_execute_inner(
            job_config,
            config_metadata,
            file_lists,
            **kwargs,
        )
    finally:
        # Always tidy-up job statuses
        completed_jobs[this_id] = running_jobs[this_id]
        del running_jobs[this_id]
    return results



import pathlib

from runem.job_runner_simple_command import (
    job_runner_simple_command,
    validate_simple_command,
)
from runem.job_wrapper_python import get_job_wrapper_py_func
from runem.types.runem_config import JobWrapper
from runem.types.types_jobs import JobFunction


def get_job_wrapper(job_wrapper: JobWrapper, cfg_filepath: pathlib.Path) -> JobFunction:
    """Given a job-description determines the job-runner, returning it as a function.

    NOTE: Side-effects: also re-addressed the job-config in the case of functions see
          get_job_function.
    """
    if "command" in job_wrapper:
        # validate that the command is "understandable" and usable.
        command_string: str = job_wrapper["command"]
        validate_simple_command(command_string)
        return job_runner_simple_command

    # if we do not have a simple command address assume we have just an addressed
    # function
    return get_job_wrapper_py_func(job_wrapper, cfg_filepath)



import pathlib

from typing_extensions import Unpack

from runem.run_command import run_command
from runem.types import FilePathList, JobKwargs


def _json_validate(
    **kwargs: Unpack[JobKwargs],
) -> None:
    label = kwargs["label"]
    json_files: FilePathList = kwargs["file_list"]
    json_with_comments = (
        "cspell.json",
        "tsconfig.spec.json",
        "launch.json",
        "settings.json",
    )
    for json_file in json_files:
        json_path = pathlib.Path(json_file)
        if not json_path.exists():
            raise RuntimeError(
                f"could not find '{str(json_path)}, in {pathlib.Path('.').absolute()}"
            )
        if json_path.name in json_with_comments:
            # until we use a validator that allows comments in json, skip these
            continue

        cmd = ["python", "-m", "json.tool", f"{json_file}"]
        kwargs["label"] = f"{label} {json_file}"
        run_command(cmd=cmd, **kwargs)



#!/usr/bin/env python3
"""`runem`, runs Lursight's dev-ops tools, hopefully as fast as possible.

We don't yet:
- account for load
- check for diffs in code to only test changed code
- do any git-related stuff, like:
  - compare head to merge-target branch
  - check for changed files
- support non-git repos
- not stream stdout to terminal
- have inter-job dependencies as that requires a smarter scheduler, we workaround
  this with phases, for now

We do:
- use git ls-files
- run as many jobs as possible
- hope that resources are enough i.e. we DO NOT measure resource use, yet.
- time tests and tell you what used the most time, and how much time run-tests saved
  you
"""

import contextlib
import multiprocessing
import os
import pathlib
import sys
import time
import typing
from collections import defaultdict
from datetime import timedelta
from itertools import repeat
from multiprocessing.managers import DictProxy, ValueProxy
from timeit import default_timer as timer

from rich.spinner import Spinner
from rich.status import Status
from rich.text import Text

from runem.blocking_print import RICH_CONSOLE
from runem.command_line import error_on_log_logic, parse_args
from runem.config import load_project_config, load_user_configs
from runem.config_metadata import ConfigMetadata
from runem.config_parse import load_config_metadata
from runem.files import find_files
from runem.job_execute import job_execute
from runem.job_filter import filter_jobs
from runem.log import error, log, warn
from runem.report import report_on_run
from runem.run_command import RunemJobError
from runem.types.common import OrderedPhases, PhaseName
from runem.types.errors import SystemExitBad
from runem.types.filters import FilePathListLookup
from runem.types.hooks import HookName
from runem.types.runem_config import Config, Jobs, PhaseGroupedJobs
from runem.types.types_jobs import (
    JobReturn,
    JobRunMetadata,
    JobRunMetadatasByPhase,
    JobTiming,
)
from runem.utils import printable_set_coloured


def _determine_run_parameters(argv: typing.List[str]) -> ConfigMetadata:
    """Loads config, parsing cli input and produces the run config.

    This is where the power of run'em resides. We match a declarative config with useful
    command-line switches to make choosing which jobs to run fast and intuitive.

    Return a ConfigMetadata object with all the required information.
    """
    # Because we want to be able to show logging whilst parsing .runem.yml config, we
    # need to check the state of the logging-verbosity switches here, manually, as well.
    verbose = "--verbose" in argv
    silent = ("--silent" in argv) or ("-s" in argv)
    error_on_log_logic(verbose, silent)

    config: Config
    cfg_filepath: pathlib.Path
    config, cfg_filepath = load_project_config()
    user_configs: typing.List[typing.Tuple[Config, pathlib.Path]] = load_user_configs()
    config_metadata: ConfigMetadata = load_config_metadata(
        config, cfg_filepath, user_configs, silent, verbose=("--verbose" in argv)
    )

    # Now we parse the cli arguments extending them with information from the
    # .runem.yml config.
    config_metadata = parse_args(config_metadata, argv)

    if config_metadata.args.verbose:
        log(f"loaded config from {cfg_filepath}")

    return config_metadata


def _update_progress(
    phase: str,
    running_jobs: typing.Dict[str, str],
    completed_jobs: typing.Dict[str, str],
    all_jobs: Jobs,
    is_running: ValueProxy[bool],
    num_workers: int,
    show_spinner: bool,
) -> None:
    """Updates progress report periodically for running tasks.

    Args:
        phase (str): The currently running phase.
        running_jobs (Dict[str, str]): The currently running jobs.
        completed_jobs (Dict[str, str]): The jobs that have finished work.
        all_jobs (Jobs): All jobs, encompassing both completed and running jobs.
        is_running (ValueProxy[bool]): Flag indicating if jobs are still running.
        num_workers (int): Indicates the number of workers performing the jobs.
        show_spinner (bool): Whether to show the animated spinner or not.
    """
    last_running_jobs_set: typing.Set[str] = set()

    # Using the `rich` module to show a loading spinner on console
    spinner_ctx: typing.Union[Status, typing.ContextManager[None]] = (
        RICH_CONSOLE.status(Spinner("dots", text="Starting tasks..."))
        if show_spinner
        else contextlib.nullcontext()
    )

    with spinner_ctx:
        while is_running.value:
            running_jobs_set: typing.Set[str] = set(running_jobs.values())

            # Progress report
            progress: str = f"{len(completed_jobs)}/{len(all_jobs)}"
            running_jobs_list = printable_set_coloured(
                running_jobs_set,
                "blue",
            )  # Reflect current running jobs accurately
            report: str = (
                f"[green]{phase}[/green]: {progress}({num_workers}): "
                f"{running_jobs_list}"
            )
            if show_spinner:
                assert isinstance(spinner_ctx, Status)
                spinner_ctx.update(Text.from_markup(report))
            else:
                if last_running_jobs_set != running_jobs_set:
                    RICH_CONSOLE.log(report)
                    last_running_jobs_set = running_jobs_set

            # Sleep for reduced CPU usage
            time.sleep(0.1)


def _process_jobs(
    config_metadata: ConfigMetadata,
    file_lists: FilePathListLookup,
    in_out_job_run_metadatas: JobRunMetadatasByPhase,
    phase: PhaseName,
    jobs: Jobs,
    show_spinner: bool,
) -> typing.Optional[RunemJobError]:
    """Execute each given job asynchronously.

    This is where the major real-world time savings happen, and it could be
    better, much, much better.

    TODO: this is where we do the scheduling, if we wanted to be smarter about
          it and, for instance, run the longest-running job first with quicker
          jobs completing around it, then we would work out that schedule here.

    returns the exception if the any of sub-procs fails, None otherwise
    """
    max_num_concurrent_procs: int = (
        config_metadata.args.procs
        if config_metadata.args.procs != -1
        else multiprocessing.cpu_count()
    )
    num_concurrent_procs: int = min(max_num_concurrent_procs, len(jobs))
    log(
        (
            f"Running '[green]{phase}[/green]' with {num_concurrent_procs} workers (of "
            f"{max_num_concurrent_procs} max) processing {len(jobs)} jobs"
        )
    )

    subprocess_error: typing.Optional[RunemJobError] = None

    with multiprocessing.Manager() as manager:
        running_jobs: DictProxy[typing.Any, typing.Any] = manager.dict()
        completed_jobs: DictProxy[typing.Any, typing.Any] = manager.dict()
        is_running: ValueProxy[bool] = manager.Value("b", True)

        terminal_writer_process = multiprocessing.Process(
            target=_update_progress,
            args=(
                phase,
                running_jobs,
                completed_jobs,
                jobs,
                is_running,
                num_concurrent_procs,
                show_spinner,
            ),
        )
        terminal_writer_process.start()

        try:
            with multiprocessing.Pool(processes=num_concurrent_procs) as pool:
                # use starmap so we can pass down the job-configs and the args and the files
                in_out_job_run_metadatas[phase] = pool.starmap(
                    job_execute,  # no kwargs passed for jobs here
                    zip(
                        jobs,
                        repeat(running_jobs),
                        repeat(completed_jobs),
                        repeat(config_metadata),
                        repeat(file_lists),
                    ),
                )
        except RunemJobError as err:  # pylint: disable=broad-exception-caught
            subprocess_error = err
        finally:
            # Signal the terminal_writer process to exit
            is_running.value = False
            terminal_writer_process.join()

    return subprocess_error


def _process_jobs_by_phase(
    config_metadata: ConfigMetadata,
    file_lists: FilePathListLookup,
    filtered_jobs_by_phase: PhaseGroupedJobs,
    in_out_job_run_metadatas: JobRunMetadatasByPhase,
    show_spinner: bool,
) -> typing.Optional[RunemJobError]:
    """Execute each job asynchronously, grouped by phase.

    Whilst it is conceptually useful to group jobs by 'phase', Phases are
    ostensibly a poor-man's dependency graph. With a proper dependency graph
    Phases could be phased out, or at least used less. For new users, and to get
    a quick and dirty solution up and running, Phases are probably a very good
    idea and easy to grasp.

    TODO: augment (NOT REPLACE) with dependency graph. New users and hacker
          dev-ops/SREs find phases useful and, more importantly, quick to
          implement.

    returns the exception, if any thrown during run.
    """
    for phase in config_metadata.phases:
        jobs = filtered_jobs_by_phase[phase]
        if not jobs:
            # As previously reported, no jobs for this phase
            continue

        if config_metadata.args.verbose:
            log(f"Running Phase {phase}")

        failure_exception: typing.Optional[RunemJobError] = _process_jobs(
            config_metadata,
            file_lists,
            in_out_job_run_metadatas,
            phase,
            jobs,
            show_spinner,
        )
        if failure_exception is not None:
            if config_metadata.args.verbose:
                error(f"running phase {phase}: aborting run")
            return failure_exception

    # ALl phases completed aok.
    return None


MainReturnType = typing.Tuple[
    ConfigMetadata, JobRunMetadatasByPhase, typing.Optional[RunemJobError]
]


def _main(
    argv: typing.List[str],
) -> MainReturnType:
    start = timer()

    config_metadata: ConfigMetadata = _determine_run_parameters(argv)

    # first anchor the cwd to the config-file, so that git ls-files works
    os.chdir(config_metadata.cfg_filepath.parent)

    file_lists: FilePathListLookup = find_files(config_metadata)
    if not file_lists:
        warn("no files found")
        return (config_metadata, {}, None)

    if config_metadata.args.verbose:
        log(f"found {len(file_lists)} batches, ", end="")
        for tag in sorted(file_lists.keys()):
            file_list = file_lists[tag]
            log(f"{len(file_list)} '{tag}' files, ", prefix=False, end="")
        log(prefix=False)  # new line

    filtered_jobs_by_phase: PhaseGroupedJobs = filter_jobs(
        config_metadata=config_metadata,
    )
    end = timer()

    job_run_metadatas: JobRunMetadatasByPhase = defaultdict(list)
    pre_build_time: JobTiming = {
        "job": ("pre-build", (timedelta(seconds=end - start))),
        "commands": [],
    }
    job_run_metadatas["_app"].append((pre_build_time, None))

    start = timer()

    failure_exception: typing.Optional[RunemJobError] = _process_jobs_by_phase(
        config_metadata,
        file_lists,
        filtered_jobs_by_phase,
        job_run_metadatas,
        show_spinner=config_metadata.args.show_spinner,
    )

    end = timer()

    phase_run_timing: JobTiming = {
        "job": ("run-phases", timedelta(seconds=end - start)),
        "commands": [],
    }
    phase_run_report: JobReturn = None
    phase_run_metadata: JobRunMetadata = (phase_run_timing, phase_run_report)
    job_run_metadatas["_app"].append(phase_run_metadata)
    return config_metadata, job_run_metadatas, failure_exception


def timed_main(argv: typing.List[str]) -> None:
    """A main-entry point that runs the application reports on it.

    IMPORTANT: this should remain a lightweight wrapper around _main() so that timings
               are representative.
    """
    start = timer()
    config_metadata: ConfigMetadata
    job_run_metadatas: JobRunMetadatasByPhase
    failure_exception: typing.Optional[RunemJobError]
    config_metadata, job_run_metadatas, failure_exception = _main(argv)
    phase_run_oder: OrderedPhases = config_metadata.phases
    end = timer()
    time_taken: timedelta = timedelta(seconds=end - start)
    wall_clock_time_saved: timedelta
    system_time_spent: timedelta
    system_time_spent, wall_clock_time_saved = report_on_run(
        phase_run_oder, job_run_metadatas, time_taken
    )
    message: str = "[green bold]DONE[/green bold]: runem took"
    if failure_exception:
        message = "[red bold]FAILED[/red bold]: your jobs failed after"
    log(
        (
            f"{message}: {time_taken.total_seconds()}s, "
            f"saving you [green]{wall_clock_time_saved.total_seconds()}s[/green], "
            "without runem you would have waited "
            f"[red]{system_time_spent.total_seconds()}s[/red]"
        )
    )

    config_metadata.hook_manager.invoke_hooks(
        hook_name=HookName.ON_EXIT,
        config_metadata=config_metadata,
        wall_clock_time_saved=wall_clock_time_saved,
    )

    if failure_exception is not None:
        # we got a failure somewhere, now that we've reported the timings we
        # re-raise.
        error(failure_exception.stdout)
        raise SystemExitBad(1) from failure_exception


if __name__ == "__main__":
    timed_main(sys.argv)



import re
import typing

from runem.report import replace_bar_graph_characters


def sanitise_reports_footer(stdout: str) -> typing.List[str]:
    """Strips variable content like floats and bar-graphs from the std out."""
    special_char: str = "="
    bar_less_stdout: str = replace_bar_graph_characters(
        stdout,
        end_str=" ",  # strip all lines
        replace_char=special_char,  # use a char that isn't used elsewise
    ).replace(special_char, "")

    lines: typing.List[str] = bar_less_stdout.split("\n")
    stripped_of_floats: typing.List[str] = [
        re.sub(r"-?\b\d+\.\d+s?\b", "<float>", text).rstrip() for text in lines
    ]
    return stripped_of_floats



import os
import pathlib
import typing
from datetime import timedelta
from subprocess import PIPE as SUBPROCESS_PIPE
from subprocess import STDOUT as SUBPROCESS_STDOUT
from subprocess import Popen
from timeit import default_timer as timer

from rich.markup import escape

from runem.log import log

TERMINAL_WIDTH = 86


class RunemJobError(RuntimeError):
    """An exception type that stores the stdout/stderr.

    Designed so that we do not print the full stdout via the exception stack, instead,
    allows an opportunity to parse the markup in it.
    """

    def __init__(self, friendly_message: str, stdout: str):
        self.stdout = stdout
        super().__init__(friendly_message)


class RunCommandBadExitCode(RunemJobError):
    def __init__(self, stdout: str):
        super().__init__(friendly_message="Bad exit-code", stdout=stdout)


class RunCommandUnhandledError(RunemJobError):
    def __init__(self, stdout: str):
        super().__init__(friendly_message="Unhandled job error", stdout=stdout)


# A function type for recording timing information.
RecordSubJobTimeType = typing.Callable[[str, timedelta], None]


def parse_stdout(stdout: str, prefix: str) -> str:
    """Prefixes each line of output with a given label, except trailing new lines."""
    # Edge case: Return the prefix immediately for an empty string
    if not stdout:
        return prefix

    # Stop errors in `rich` by parsing out anything that might look like
    # rich-markup.
    stdout = escape(stdout)

    # Split stdout into lines
    lines = stdout.split("\n")

    # Apply prefix to all lines except the last if it's empty (due to a trailing newline)
    modified_lines = [f"{prefix}{escape(line)}" for line in lines[:-1]] + (
        [f"{prefix}{escape(lines[-1])}"]
    )

    # Join the lines back together, appropriately handling the final newline
    modified_stdout: str = "\n".join(modified_lines)

    return modified_stdout


def _prepare_environment(
    env_overrides: typing.Optional[typing.Dict[str, str]],
) -> typing.Dict[str, str]:
    """Returns a consolidated environment merging os.environ and overrides."""
    # first and always, copy in the environment
    run_env: typing.Dict[str, str] = {
        **os.environ,  # copy in the environment
    }
    if env_overrides:
        # overload the os.environ with overrides
        run_env.update(env_overrides)
    return run_env


def _log_command_execution(
    cmd_string: str,
    label: str,
    env_overrides: typing.Optional[typing.Dict[str, str]],
    valid_exit_ids: typing.Optional[typing.Tuple[int, ...]],
    decorate_logs: bool,
    verbose: bool,
    cwd: typing.Optional[pathlib.Path] = None,
) -> None:
    """Logs out useful debug information on '--verbose'."""
    if verbose:
        log(
            f"running: start: [blue]{label}[/blue]: [yellow]{cmd_string}[yellow]",
            prefix=decorate_logs,
        )
        if valid_exit_ids is not None:
            valid_exit_strs = ",".join(str(exit_code) for exit_code in valid_exit_ids)
            log(
                f"\tallowed return ids are: [green]{valid_exit_strs}[/green]",
                prefix=decorate_logs,
            )

        if env_overrides:
            env_overrides_as_string = " ".join(
                [f"{key}='{value}'" for key, value in env_overrides.items()]
            )
            log(
                f"ENV OVERRIDES: [yellow]{env_overrides_as_string} {cmd_string}[/yellow]",
                prefix=decorate_logs,
            )

        if cwd:
            log(f"cwd: {str(cwd)}", prefix=decorate_logs)


def run_command(  # noqa: C901
    cmd: typing.List[str],  # 'cmd' is the only thing that can't be optionally kwargs
    label: str,
    verbose: bool,
    env_overrides: typing.Optional[typing.Dict[str, str]] = None,
    ignore_fails: bool = False,
    valid_exit_ids: typing.Optional[typing.Tuple[int, ...]] = None,
    cwd: typing.Optional[pathlib.Path] = None,
    record_sub_job_time: typing.Optional[RecordSubJobTimeType] = None,
    decorate_logs: bool = True,
    **kwargs: typing.Any,
) -> str:
    """Runs the given command, returning stdout or throwing on any error."""
    cmd_string = " ".join(cmd)

    if record_sub_job_time is not None:
        # start the capture of how long this sub-task takes.
        start = timer()

    run_env: typing.Dict[str, str] = _prepare_environment(
        env_overrides,
    )
    _log_command_execution(
        cmd_string,
        label,
        env_overrides,
        valid_exit_ids,
        decorate_logs,
        verbose,
        cwd,
    )

    if valid_exit_ids is None:
        valid_exit_ids = (0,)

    # init the process in case it throws for things like not being able to
    # convert the command to a list of strings.
    process: typing.Optional[Popen[str]] = None
    stdout: str = ""
    try:
        with Popen(
            cmd,
            env=run_env,
            stdout=SUBPROCESS_PIPE,
            stderr=SUBPROCESS_STDOUT,
            cwd=cwd,
            text=True,
            bufsize=1,  # buffer it for every character return
            universal_newlines=True,
        ) as process:
            # Read output line by line as it becomes available
            assert process.stdout is not None
            for line in process.stdout:
                stdout += line
                if verbose:
                    # print each line of output, assuming that each has a newline
                    log(
                        parse_stdout(
                            line, prefix=f"[green]| [/green][blue]{label}[/blue]: "
                        ),
                        prefix=False,
                    )

            # Wait for the subprocess to finish and get the exit code
            process.wait()

            if process.returncode not in valid_exit_ids:
                valid_exit_strs = ",".join(
                    [str(exit_code) for exit_code in valid_exit_ids]
                )
                raise RunCommandBadExitCode(
                    (
                        f"non-zero exit [red]{process.returncode}[/red] (allowed are "
                        f"[green]{valid_exit_strs}[/green]) from {cmd_string}"
                    )
                )
    except BaseException as err:
        if ignore_fails:
            return ""
        parsed_stdout: str = (
            parse_stdout(stdout, prefix="[red]| [/red]") if process else ""
        )
        env_overrides_as_string = ""
        if env_overrides:
            env_overrides_as_string = " ".join(
                [f"{key}='{value}'" for key, value in env_overrides.items()]
            )
            env_overrides_as_string = f"{env_overrides_as_string} "
        error_string = (
            f"runem: [red bold]FATAL[/red bold]: command failed: [blue]{label}[/blue]"
            f"\n\t[yellow]{env_overrides_as_string}{cmd_string}[/yellow]"
            f"\n[red underline]| ERROR[/red underline]: [blue]{label}[/blue]"
            f"\n{str(parsed_stdout)}"
            f"\n[red underline]| ERROR END[/red underline]"
        )

        if isinstance(err, RunCommandBadExitCode):
            raise RunCommandBadExitCode(error_string) from err
        # fallback to raising a RunCommandUnhandledError
        raise RunCommandUnhandledError(error_string) from err

    if verbose:
        log(
            f"running: done: [blue]{label}[/blue]: [yellow]{cmd_string}[/yellow]",
            prefix=decorate_logs,
        )

    if record_sub_job_time is not None:
        # Capture how long this run took
        end = timer()
        time_taken: timedelta = timedelta(seconds=end - start)
        record_sub_job_time(label, time_taken)

    return stdout



from runem.types.common import FilePathList, JobName
from runem.types.options import Options
from runem.types.types_jobs import HookKwargs, JobKwargs, JobReturn, JobReturnData

__all__ = [
    "FilePathList",
    "HookKwargs",
    "JobName",
    "JobReturn",
    "JobReturnData",
    "Options",
    "JobKwargs",
]



import shlex
import typing

from typing_extensions import Unpack

from runem.config_metadata import ConfigMetadata
from runem.run_command import run_command
from runem.types.common import FilePathList
from runem.types.options import OptionsWritable
from runem.types.runem_config import JobConfig
from runem.types.types_jobs import AllKwargs


def validate_simple_command(command_string: str) -> typing.List[str]:
    """Use shlex to handle parsing of the command string, a non-trivial problem."""
    split_command: typing.List[str] = shlex.split(command_string)
    return split_command


def job_runner_simple_command(
    **kwargs: Unpack[AllKwargs],
) -> None:
    """Parses the command and tries to run it via the system.

    Commands inherit the environment.
    """
    # assume we have the job.command entry, allowing KeyError to propagate up
    job_config: JobConfig = kwargs["job"]
    command_string: str = job_config["command"]

    command_string_files: str = command_string
    if "{file_list}" in command_string:
        file_list: FilePathList = kwargs["file_list"]
        file_list_with_quotes: typing.List[str] = [
            f'"{str(file_path)}"' for file_path in file_list
        ]
        command_string_files = command_string.replace(
            "{file_list}", " ".join(file_list_with_quotes)
        )

    config_metadata: ConfigMetadata = kwargs["config_metadata"]
    options: OptionsWritable = config_metadata.options
    command_string_options: str = command_string_files
    for name, value in options.items():
        # For now, just pass `--option-name`, `--check` or similar to the
        # command line. At some point we will want this to be cleverer, but
        # this will do for now.
        option_search = f"{{{name}}}"
        if option_search in command_string_files:
            replacement = ""
            if value:
                replacement = f"--{name}"
            command_string_options = command_string_options.replace(
                option_search, replacement
            )

    # use shlex to handle parsing of the command string, a non-trivial problem.
    cmd = validate_simple_command(command_string_options)

    # preserve quotes for consistent handling of strings and avoid the "word
    # splitting" problem for unix-like shells.
    cmd_with_quotes = [f'"{token}"' if " " in token else token for token in cmd]

    run_command(cmd=cmd_with_quotes, **kwargs)



import typing
from unittest.mock import MagicMock, call, patch

import pytest

from runem import blocking_print


@pytest.fixture(name="mock_print")
def mock_print_fixture() -> typing.Generator[MagicMock, None, None]:
    with patch("runem.blocking_print.RICH_CONSOLE.print") as mock_print:
        yield mock_print


@pytest.fixture(name="mock_sleep")
def mock_sleep_fixture() -> typing.Generator[MagicMock, None, None]:
    with patch("runem.blocking_print.time.sleep") as mock_sleep:
        yield mock_sleep


def test_blocking_print_success_first_try(
    mock_print: MagicMock, mock_sleep: MagicMock
) -> None:
    """Test that blocking_print prints the message successfully on the first try."""
    blocking_print.blocking_print("Test message")
    mock_print.assert_called_once_with("Test message", end="\n")
    mock_sleep.assert_not_called()


def test_blocking_print_retries_and_succeeds(
    mock_print: MagicMock, mock_sleep: MagicMock
) -> None:
    """Test that blocking_print retries on BlockingIOError and succeeds."""
    # Configure print to raise BlockingIOError twice before succeeding
    mock_print.side_effect = [BlockingIOError, BlockingIOError, None]
    blocking_print.blocking_print("Test message", end="!")
    assert mock_print.call_count == 3
    # Check that print was called with the correct arguments each time
    mock_print.assert_has_calls([call("Test message", end="!")] * 3)
    # Ensure sleep was called twice
    assert mock_sleep.call_count == 2


def test_blocking_print_exhausts_retries(
    mock_print: MagicMock, mock_sleep: MagicMock
) -> None:
    """Test that blocking_print exhausts retries and fails to print."""
    # Configure print to always raise BlockingIOError
    mock_print.side_effect = BlockingIOError
    blocking_print.blocking_print("Test message", max_retries=3, sleep_time_s=0.1)
    assert mock_print.call_count == 3
    assert mock_sleep.call_count == 3


def test_blocking_print_empty_message(
    mock_print: MagicMock, mock_sleep: MagicMock
) -> None:
    """Test that blocking_print handles an empty message."""
    blocking_print.blocking_print()
    mock_print.assert_called_once_with("", end="\n")
    mock_sleep.assert_not_called()


def test_blocking_print_custom_end_parameter(
    mock_print: MagicMock, mock_sleep: MagicMock
) -> None:
    """Optional: Test handling of non-default end parameter"""
    blocking_print.blocking_print("Test", end="\n")
    mock_print.assert_called_once_with("Test", end="\n")
    mock_sleep.assert_not_called()



from unittest.mock import Mock, patch

import runem.cli


@patch(
    "runem.cli.timed_main",
    return_value=None,
)
def test_main(patched_main: Mock) -> None:
    runem.cli.main()
    patched_main.assert_called_once()



import pathlib
import shutil
from unittest.mock import Mock, patch

import pytest

from runem.job_wrapper_python import (
    _load_python_function_from_module,
    get_job_wrapper_py_func,
)
from runem.types.errors import FunctionNotFound
from runem.types.runem_config import JobConfig
from runem.types.types_jobs import JobFunction


def test_get_job_function(tmp_path: pathlib.Path) -> None:
    """Tests that loading a function works.

    The files have to be in the same path and we use the tmp_path so we copy the right
    files there.
    """
    source_file = pathlib.Path(__file__)
    copied_py = tmp_path / source_file.name
    shutil.copyfile(source_file, copied_py)

    job_config: JobConfig = {
        "addr": {
            "file": str(copied_py),
            "function": "test_get_job_function",
        },
        "label": "reformat py",
        "when": {
            "phase": "edit",
            "tags": set(
                (
                    "py",
                    "format",
                )
            ),
        },
    }
    loaded_func: JobFunction = get_job_wrapper_py_func(
        job_config, tmp_path / ".runem.yml"
    )
    assert loaded_func is not None
    assert loaded_func.__name__ == "test_get_job_function"


def test_get_job_function_handles_missing_function(tmp_path: pathlib.Path) -> None:
    """Tests that loading a non-existent function in a valid file fails gracefully.

    The files have to be in the same path and we use the tmp_path so we copy the right
    files there.
    """
    source_file = pathlib.Path(__file__)
    copied_py = tmp_path / source_file.name
    shutil.copyfile(source_file, copied_py)

    job_config: JobConfig = {
        "addr": {
            "file": str(copied_py),
            "function": "function_does_not_exist",
        },
        "label": "reformat py",
        "when": {
            "phase": "edit",
            "tags": set(
                (
                    "py",
                    "format",
                )
            ),
        },
    }

    with pytest.raises(FunctionNotFound):
        # this should throw an exception
        get_job_wrapper_py_func(job_config, tmp_path / ".runem.yml")


def test_get_job_function_handles_missing_module(tmp_path: pathlib.Path) -> None:
    """Tests that loading a non-existent function in a valid file fails gracefully.

    The files have to be in the same path and we use the tmp_path so we copy the right
    files there.
    """
    source_file = pathlib.Path(__file__)
    not_copied_py = tmp_path / source_file.name

    job_config: JobConfig = {
        "addr": {
            "file": str(not_copied_py),
            "function": "function_does_not_exist",
        },
        "label": "reformat py",
        "when": {
            "phase": "edit",
            "tags": set(
                (
                    "py",
                    "format",
                )
            ),
        },
    }

    with pytest.raises(FunctionNotFound):
        # this should throw an exception
        get_job_wrapper_py_func(job_config, tmp_path / ".runem.yml")


@patch(
    "runem.job_wrapper_python.module_spec_from_file_location",
    return_value=None,
)
def test_load_python_function_from_module_handles_module_spec_loading(
    mock_spec_from_file_location: Mock,
) -> None:
    """Tests that the importlib internals failing to load a module-spec is handled.

    mocks importlib.util.spec_from_file_location to return None
    """
    file_path: pathlib.Path = pathlib.Path(__file__)
    base_path = file_path.parent
    with pytest.raises(FunctionNotFound) as err_info:
        _load_python_function_from_module(
            base_path / ".runem.no_exist.yml",
            "test_module_name",
            file_path,
            "test_load_python_function_from_module_handles_module_spec_loading",
        )
    assert str(err_info.value).startswith(
        (
            "unable to load "
            "'test_load_python_function_from_module_handles_module_spec_loading' from"
        )
    )
    mock_spec_from_file_location.assert_called()


@patch(
    "runem.job_wrapper_python.module_from_spec",
    return_value=None,
)
def test_load_python_function_from_module_handles_module_from_spec_failing(
    mock_module_from_spec: Mock,
) -> None:
    """Tests that another case of importlib internals failing is handled.

    mocks importlib.util.module_from_spec to return None
    """
    file_path: pathlib.Path = pathlib.Path(__file__)
    base_path = file_path.parent
    with pytest.raises(FunctionNotFound, match=("unable to load module")):
        _load_python_function_from_module(
            base_path / ".runem.no_exist.yml",
            "test_module_name",
            file_path,
            "test_load_python_function_from_module_handles_module_spec_loading",
        )
    mock_module_from_spec.assert_called()


def test_load_python_function_success(tmp_path: pathlib.Path) -> None:
    # Create a temporary Python module file
    p = tmp_path / "hello.py"
    p.write_text('def hello():\n    return "Hello, World!"')

    # Use the function to load the module and retrieve the function
    hello_func = _load_python_function_from_module(
        p, "hello_module", pathlib.Path("hello.py"), "hello"
    )

    # True if the function can be called and returns expected result
    assert callable(hello_func)
    assert hello_func() == "Hello, World!"  # type: ignore


def test_load_python_function_module_not_found(tmp_path: pathlib.Path) -> None:
    # Use the function to load a non-existing module
    with pytest.raises(FunctionNotFound):
        _load_python_function_from_module(
            tmp_path, "not_found_module", pathlib.Path("not_found.py"), "not_found"
        )


def test_load_python_function_not_found(tmp_path: pathlib.Path) -> None:
    # Create a temporary Python module file
    p = tmp_path / "hello.py"
    p.write_text(
        """
    def hello():
        return "Hello, World!"
    """
    )

    # Use the function to load the module but try to retrieve a non-existent function
    with pytest.raises(FunctionNotFound):
        _load_python_function_from_module(
            tmp_path, "hello_module", pathlib.Path("hello.py"), "not_found"
        )


def test_load_python_function_invalid_module(tmp_path: pathlib.Path) -> None:
    # Create an invalid Python module file
    p = tmp_path / "invalid.py"
    p.write_text(
        """
    def 123invalid():
        return "Invalid!"
    """
    )

    # Use the function to load the invalid module and expect a FunctionNotFound exception
    with pytest.raises(FunctionNotFound):
        _load_python_function_from_module(
            tmp_path, "invalid_module", pathlib.Path("invalid.py"), "123invalid"
        )



import io
import typing
from contextlib import redirect_stdout
from pathlib import Path
from unittest.mock import ANY, Mock, patch

from runem.config_metadata import ConfigMetadata
from runem.job import Job
from runem.job_runner_simple_command import job_runner_simple_command
from runem.types.runem_config import JobConfig
from tests.test_runem import gen_dummy_config_metadata


@patch(
    "runem.job_runner_simple_command.run_command",
    # return_value=None,
)
def test_job_runner_simple_command(mock_run_command: Mock) -> None:
    """Tests the basics of job_runner_simple_command."""
    job_config: JobConfig = {
        "command": "echo 'testing job_runner_simple_command'",
    }
    config_metadata: ConfigMetadata = gen_dummy_config_metadata()
    with io.StringIO() as buf, redirect_stdout(buf):
        ret: None = job_runner_simple_command(  # type: ignore[func-returns-value]
            config_metadata=config_metadata,
            options=config_metadata.options,  # type: ignore
            file_list=[],
            procs=config_metadata.args.procs,
            root_path=Path("."),
            verbose=True,  # config_metadata.args.verbose,
            # unpack useful data points from the job_config
            label=Job.get_job_name(job_config),
            job=job_config,
        )
        run_command_stdout = buf.getvalue()

    assert ret is None
    assert run_command_stdout.split("\n") == [""]
    mock_run_command.assert_called_once_with(
        cmd=["echo", '"testing job_runner_simple_command"'],
        config_metadata=ANY,
        file_list=[],
        job={"command": "echo 'testing job_runner_simple_command'"},
        label="echo 'testing job_runner_simple_command'",
        options=ANY,
        procs=1,
        root_path=Path("."),
        verbose=True,
    )


@patch(
    "runem.job_runner_simple_command.run_command",
    # return_value=None,
)
def test_job_runner_simple_command_with_file_list(mock_run_command: Mock) -> None:
    """Tests the basics of job_runner_simple_command."""
    test_cmd_string: str = (
        'echo "some option before files" {file_list} "some option after files"'
    )
    job_config: JobConfig = {
        "command": test_cmd_string,
    }
    config_metadata: ConfigMetadata = gen_dummy_config_metadata()
    file_list: typing.List[typing.Union[str, Path]] = [
        Path("file1"),
        "file2",
        "file with spaces",
    ]
    with io.StringIO() as buf, redirect_stdout(buf):
        ret: None = job_runner_simple_command(  # type: ignore[func-returns-value]
            file_list=file_list,  # type: ignore[arg-type] # intentional type mismatch
            job=job_config,
            label=Job.get_job_name(job_config),
            config_metadata=config_metadata,
            options=config_metadata.options,  # type: ignore
            procs=config_metadata.args.procs,
            root_path=Path("."),
            verbose=True,  # config_metadata.args.verbose,
        )
        run_command_stdout = buf.getvalue()

    assert ret is None
    assert run_command_stdout.split("\n") == [""]
    mock_run_command.assert_called_once_with(
        cmd=[
            "echo",
            '"some option before files"',
            "file1",
            "file2",
            '"file with spaces"',
            '"some option after files"',
        ],
        config_metadata=ANY,
        file_list=file_list,
        job=job_config,
        label=test_cmd_string,
        options=ANY,
        procs=1,
        root_path=Path("."),
        verbose=True,
    )


@patch(
    "runem.job_runner_simple_command.run_command",
    # return_value=None,
)
def test_job_runner_simple_command_with_option(mock_run_command: Mock) -> None:
    """Tests that option-passing to jobs, pass --option_on but not --option_off."""
    test_cmd_string: str = (
        'echo "some option before switch" {option_on} '
        '{option_off} "some option after switch"'
    )
    job_config: JobConfig = {
        "command": test_cmd_string,
    }
    config_metadata: ConfigMetadata = gen_dummy_config_metadata()
    file_list: typing.List[typing.Union[str, Path]] = [
        Path("file1"),
        "file2",
        "file with spaces",
    ]
    with io.StringIO() as buf, redirect_stdout(buf):
        ret: None = job_runner_simple_command(  # type: ignore[func-returns-value]
            file_list=file_list,  # type: ignore[arg-type] # intentional type mismatch
            job=job_config,
            label=Job.get_job_name(job_config),
            config_metadata=config_metadata,
            options=config_metadata.options,  # type: ignore
            procs=config_metadata.args.procs,
            root_path=Path("."),
            verbose=True,  # config_metadata.args.verbose,
        )
        run_command_stdout = buf.getvalue()

    assert ret is None
    assert run_command_stdout.split("\n") == [""]
    mock_run_command.assert_called_once_with(
        cmd=[
            "echo",
            '"some option before switch"',
            "--option_on",
            # Not this -> "--option_off",
            '"some option after switch"',
        ],
        config_metadata=ANY,
        file_list=file_list,
        job=job_config,
        label=test_cmd_string,
        options={"option_on": True, "option_off": False},
        procs=1,
        root_path=Path("."),
        verbose=True,
    )



"""tests/test_config_validate.py.

Unit‑tests for `runem.config_validate`.

Coverage
========
1. `_load_runem_schema`
   • success path (schema present)
   • error path  (schema missing → SystemExitBad)

2. `validate_runem_file`
   • happy path   (no validation errors)
   • failure path (≥1 validation error → SystemExit, correct logging)

Design notes
============
* All I/O is faked with `monkeypatch`; no real files created in package dirs.
* Dummy error objects stand‑in for `jsonschema.ValidationError` to keep deps thin.
* Each test asserts **behaviour**, not internals.
"""
# pylint: disable=protected-access

from __future__ import annotations

import pathlib
from typing import List, Union

import pytest

# --------------------------------------------------------------------------- #
# Test helpers
# --------------------------------------------------------------------------- #
import runem.config_validate as uut  # unit under test


class _DummyErr:  # pylint: disable=too-few-public-methods
    """Minimal substitute for jsonschema.ValidationError."""

    def __init__(self, path: List[Union[str, int]], message: str) -> None:
        self.path = path
        self.message = message


# --------------------------------------------------------------------------- #
# _load_runem_schema
# --------------------------------------------------------------------------- #


def test_load_runem_schema_ok(monkeypatch: pytest.MonkeyPatch) -> None:
    """Returns loaded schema when file exists."""
    sentinel: object = object()

    monkeypatch.setattr(
        uut.pathlib.Path,  # type: ignore[attr-defined]
        "exists",
        lambda self: True,  # noqa: ANN001
        raising=False,
    )
    monkeypatch.setattr(uut, "load_yaml_object", lambda _: sentinel)

    assert uut._load_runem_schema() is sentinel  # pyright: ignore [private‑access]


def test_load_runem_schema_missing(monkeypatch: pytest.MonkeyPatch) -> None:
    """Raises `SystemExitBad` and logs when schema file is absent."""
    calls: list[str] = []

    monkeypatch.setattr(
        uut.pathlib.Path,  # type: ignore[attr-defined]
        "exists",
        lambda self: False,  # noqa: ANN001
        raising=False,
    )
    monkeypatch.setattr(uut, "error", lambda msg: calls.append(str(msg)))

    with pytest.raises(uut.SystemExitBad):  # type: ignore[attr-defined]
        uut._load_runem_schema()

    assert calls and "schema file not found" in calls[0]


# --------------------------------------------------------------------------- #
# validate_runem_file
# --------------------------------------------------------------------------- #


def test_validate_runem_file_valid(
    monkeypatch: pytest.MonkeyPatch, tmp_path: pathlib.Path
) -> None:
    """No exception when `validate_yaml` returns ∅ errors."""
    cfg_path = tmp_path / "dummy.yml"
    cfg_path.touch()

    monkeypatch.setattr(uut, "_load_runem_schema", object, raising=True)
    monkeypatch.setattr(uut, "validate_yaml", lambda *_: [])

    # should run cleanly
    uut.validate_runem_file(cfg_path, all_config={})


def test_validate_runem_file_invalid(
    monkeypatch: pytest.MonkeyPatch, tmp_path: pathlib.Path
) -> None:
    """Logs every error then raises `SystemExit`."""
    cfg_path = tmp_path / "bad.yml"
    cfg_path.touch()

    logged: list[str] = []
    errors: list[str] = []

    monkeypatch.setattr(uut, "_load_runem_schema", object, raising=True)
    monkeypatch.setattr(
        uut,
        "validate_yaml",
        lambda *_: [
            _DummyErr(path=["options", 0, "type"], message="unknown field"),
            _DummyErr(path=[], message="root issue"),
        ],
    )
    monkeypatch.setattr(uut, "error", lambda msg: errors.append(str(msg)))
    monkeypatch.setattr(uut, "log", lambda msg: logged.append(str(msg)))

    with pytest.raises(SystemExit):
        uut.validate_runem_file(cfg_path, all_config={})

    # one high‑level error + per‑item log entries
    assert errors[0].startswith("failed to validate runem config")
    assert len(logged) == 2



import pathlib

from typing_extensions import Unpack

from runem.run_command import run_command
from runem.types import JobKwargs, Options


def _job_yarn_deps(
    **kwargs: Unpack[JobKwargs],
) -> None:
    """Installs the yarn deps."""
    options: Options = kwargs["options"]

    install_requested = options["install-deps"]
    if not (install_requested):
        root_path: pathlib.Path = kwargs["root_path"]
        if (root_path / "node_modules").exists():
            # An install was not requested, nor required.
            return

    install_cmd = [
        "yarn",
        "install",
    ]

    run_command(cmd=install_cmd, **kwargs)


def _job_prettier(
    **kwargs: Unpack[JobKwargs],
) -> None:
    """Runs prettifier on files, including json and maybe yml file.

    TODO: connect me up!
    """
    options: Options = kwargs["options"]
    command_variant = "pretty"
    if options["check-only"]:
        command_variant = "prettyCheck"

    pretty_cmd = [
        "yarn",
        "run",
        command_variant,
    ]

    run_command(cmd=pretty_cmd, **kwargs)



import typing
from collections import defaultdict

import pytest

from runem.job import BadWhenConfigLocation, Job, NoJobName
from runem.types.common import FilePathList, JobTags
from runem.types.filters import FilePathListLookup
from runem.types.runem_config import JobConfig


@pytest.mark.parametrize(
    "job_config, expected_result",
    [
        ({"when": {}}, None),
        ({"when": {"tags": []}}, set()),
        ({"when": {"tags": ["tag1"]}}, {"tag1"}),
        ({"when": {"tags": ["tag1", "tag2", "tag3"]}}, {"tag1", "tag2", "tag3"}),
        ({}, None),
        ({"when": {}}, None),
    ],
)
def test_get_job_tags(
    job_config: JobConfig, expected_result: typing.Optional[JobTags]
) -> None:
    result: typing.Optional[JobTags] = Job.get_job_tags(job_config)
    assert result == expected_result


def test_get_job_tags_bad_tags_path() -> None:
    """Tests that the 'tags' entry is on 'when' and not on root."""
    job_config: JobConfig = {  # type: ignore[typeddict-unknown-key]
        "tags": ["dummy tags"],
    }
    with pytest.raises(BadWhenConfigLocation):
        Job.get_job_tags(job_config)


def test_get_job_tags_bad_phase_path() -> None:
    """Tests that the 'phase' entry is on 'when' and not on root."""
    job_config: JobConfig = {  # type: ignore[typeddict-unknown-key]
        "phase": "dummy tags",
    }
    with pytest.raises(BadWhenConfigLocation):
        Job.get_job_tags(job_config)


@pytest.fixture(name="file_lists")
def file_lists_fixture() -> FilePathListLookup:
    """Define a sample file_lists dictionary for testing."""
    file_lists: FilePathListLookup = defaultdict(list)
    file_lists.update(
        {
            "tag1": ["file1.txt", "file2.txt"],
            "tag2": ["file3.txt", "file4.txt"],
            "tag3": ["file5.txt"],
        }
    )
    return file_lists


@pytest.mark.parametrize(
    "job_tags, expected_result",
    [
        (None, ["file1.txt", "file2.txt", "file3.txt", "file4.txt", "file5.txt"]),
        ({"tag1"}, ["file1.txt", "file2.txt"]),
        ({"tag2", "tag3"}, ["file3.txt", "file4.txt", "file5.txt"]),
        ({"tag4"}, []),
    ],
)
def test_get_job_files(
    file_lists: FilePathListLookup,
    job_tags: typing.Optional[JobTags],
    expected_result: FilePathList,
) -> None:
    """Test cases for the get_job_files method."""
    # Call the method under test
    result: FilePathList = Job.get_job_files(file_lists, job_tags)

    # Assert the result matches the expected outcome
    assert result == sorted(expected_result)


def test_get_job_name() -> None:
    """Test case for the get_job_name method."""
    # Define a sample job configuration for testing
    job_config: JobConfig = {
        "label": "Job 1",
        "command": "python script.py",
        "addr": {"file": "script.py", "function": "main"},
    }

    # Call the method under test
    result: str = Job.get_job_name(job_config)

    # Assert the result matches the expected outcome
    assert result == "Job 1"


def test_get_job_name_command_key() -> None:
    """Test case for the get_job_name method when using the "command" key."""
    job_config: JobConfig = {
        "command": "python script.py",
    }
    result: str = Job.get_job_name(job_config)
    assert result == "python script.py"


def test_get_job_name_addr_key() -> None:
    """Test case for the get_job_name method when using the "addr" key."""
    job_config: JobConfig = {
        "addr": {"file": "script.py", "function": "main"},
    }
    result: str = Job.get_job_name(job_config)
    assert result == "script.py.main"


def test_get_job_name_invalid_config() -> None:
    """Test case for the get_job_name method with an invalid configuration."""
    job_config: JobConfig = {}
    with pytest.raises(NoJobName):
        Job.get_job_name(job_config)



import pathlib
from unittest.mock import Mock, patch

from runem import job_runner_simple_command
from runem.job_wrapper import get_job_wrapper
from runem.types.runem_config import JobConfig
from runem.types.types_jobs import JobFunction

# use a string as a ad type to test call
DUMMY_FUNCTION: JobFunction = "intentionally bad type"  # type: ignore[assignment]


@patch(
    "runem.job_wrapper.get_job_wrapper_py_func",
    return_value=DUMMY_FUNCTION,
)
def test_get_job_wrapper_python_wrapper_job(mock_get_job_wrapper_py_func: Mock) -> None:
    """Checks that the python wrapper job lookup is called."""
    job_config: JobConfig = {
        "addr": {
            "file": __file__,
            "function": "test_get_job_wrapper",
        },
        "label": "reformat py",
        "when": {
            "phase": "edit",
            "tags": set(
                (
                    "py",
                    "format",
                )
            ),
        },
    }
    func_obj: JobFunction = get_job_wrapper(
        job_config, cfg_filepath=pathlib.Path(__file__)
    )
    assert func_obj == DUMMY_FUNCTION

    # with a python-function job-config if should call the mock once
    mock_get_job_wrapper_py_func.assert_called_once()


@patch(
    "runem.job_wrapper.get_job_wrapper_py_func",
    return_value=None,
)
def test_get_job_wrapper_simple_command_job(mock_get_job_wrapper_py_func: Mock) -> None:
    """Checks that the job_runner_simple_command is returned for 'command' jobs."""
    job_config: JobConfig = {
        "command": "echo 'testing exec'",
    }
    func_obj: JobFunction = get_job_wrapper(
        job_config, cfg_filepath=pathlib.Path(__file__)
    )

    assert func_obj.__name__ == job_runner_simple_command.__name__.split(".")[1]

    # with simple jobs the python wrapper lookup should NOT be called.
    mock_get_job_wrapper_py_func.assert_not_called()



import pathlib

from packaging.version import Version


def get_runem_version() -> Version:
    """Returns the Version object representing runem's current version."""
    return Version(
        (pathlib.Path(__file__).parent / "VERSION").read_text("utf8").strip()
    )



from typing import Optional


class FunctionNotFound(ValueError):
    """Thrown when the test-function cannot be found."""

    pass


class SystemExitBad(SystemExit):
    def __init__(self, code: Optional[int] = None) -> None:
        super().__init__()
        self.code = 1 if code is None else code  # non-zero bad exit code
        assert self.code > 0, "A bad exit code should be non-zero and >0"



from typing import Any, List

from jsonschema import Draft202012Validator, ValidationError

# For now just return the raw ValidationErrors as a list
ValidationErrors = List[ValidationError]


def validate_yaml(yaml_data: Any, schema: Any) -> ValidationErrors:
    """Validates the give yaml data against the given schema, returning any errors.

    We use more future-looking validation so that we can have richer and more
    descriptive schema.

    Params:
        instance: JSON data loaded via `load_json` or similar
        schema: schema object compatible with a Draft202012Validator

    Returns:
        ValidationErrors: a sorted list of errors in the file, empty if none found
    """
    validator = Draft202012Validator(schema)
    errors: ValidationErrors = sorted(
        validator.iter_errors(yaml_data),
        key=lambda e: e.path,
    )

    return errors



import argparse

from runem.config_metadata import ConfigMetadata
from runem.informative_dict import InformativeDict
from runem.types.options import OptionsWritable


def initialise_options(
    config_metadata: ConfigMetadata,
    args: argparse.Namespace,
) -> OptionsWritable:
    """Initialises and returns the set of options to use for this run.

    Returns the options dictionary
    """
    options: OptionsWritable = InformativeDict(
        {option["name"]: option["default"] for option in config_metadata.options_config}
    )
    if config_metadata.options_config and args.overrides_on:  # pragma: no branch
        for option_name in args.overrides_on:  # pragma: no branch
            options[option_name] = True
    if config_metadata.options_config and args.overrides_off:  # pragma: no branch
        for option_name in args.overrides_off:
            options[option_name] = False
    return options



import io
import pathlib
import typing
from argparse import Namespace
from collections import defaultdict
from contextlib import redirect_stdout
from unittest.mock import MagicMock, Mock, patch

import pytest

from runem.config_metadata import ConfigMetadata
from runem.informative_dict import InformativeDict
from runem.job_filter import _get_jobs_matching, _should_filter_out_by_tags, filter_jobs
from runem.types.common import JobTags
from runem.types.runem_config import JobConfig, PhaseGroupedJobs


@pytest.mark.parametrize(
    "verbosity",
    [
        True,
        False,
    ],
)
def test_runem_job_filters_work_with_no_tags(verbosity: bool) -> None:
    """TODO."""
    config_file_path = pathlib.Path(__file__).parent / ".runem.yml"
    expected_job: JobConfig = {
        "addr": {
            "file": "test_config_parse.py",
            "function": "test_parse_config",
        },
        "label": "dummy job label",
        "when": {
            "phase": "dummy phase 1",
            "tags": {"dummy tag 1", "dummy tag 2"},
        },
    }
    expected_jobs: PhaseGroupedJobs = defaultdict(list)
    expected_jobs["dummy phase 1"] = [
        expected_job,
    ]
    config_metadata: ConfigMetadata = ConfigMetadata(
        cfg_filepath=config_file_path,
        phases=("dummy phase 1",),
        options_config=tuple(),
        file_filters={
            # "dummy tag": {
            #     "tag": "dummy tag",
            #     "regex": ".*1.txt",  # should match just one file
            # }
        },
        hook_manager=MagicMock(),
        jobs=expected_jobs,
        all_job_names=set(("dummy job label",)),
        all_job_phases=set(("dummy phase 1",)),
        all_job_tags=set(),
    )
    config_metadata.set_cli_data(
        args=Namespace(verbose=verbosity, procs=1),
        jobs_to_run=set(("dummy job label",)),  # JobNames,
        phases_to_run=set(),  # ignored JobPhases,
        tags_to_run=set(),  # ignored JobTags,
        tags_to_avoid=set(),  # ignored  JobTags,
        options=InformativeDict({}),  # Options,
    )
    with io.StringIO() as buf, redirect_stdout(buf):
        filter_jobs(config_metadata)
        run_command_stdout = buf.getvalue()
    if verbosity:
        assert run_command_stdout.split("\n") == [
            "runem: skipping phase 'dummy phase 1'",
            "",
        ]
    else:
        assert run_command_stdout == ""


@pytest.mark.parametrize(
    "verbosity",  # Parameter for the test function; iterates over True and False.
    [
        True,
        False,
    ],
)
def test_should_filter_out_by_tags_with_tags_to_avoid(verbosity: bool) -> None:
    """Tests we correctly filters out jobs based on tags to avoid.

    Parameters:
    - verbosity: A boolean indicating whether detailed logging is enabled.

    The test validates two conditions:
    1. That the job is correctly identified to be filtered out based on the
       presence of avoided tags.
    2. That the verbosity of the output matches the expected verbosity level.
    """
    # Set up a job configuration with specific tags.
    job: JobConfig = {
        "label": "Job1",  # Job identifier.
        "when": {
            "tags": {
                "tag1",
                "tag2",
            }
        },
    }
    tags: JobTags = {"tag1"}  # Tags present in the job.
    tags_to_avoid: JobTags = {
        "tag1",
        "tag2",
    }  # Tags that should cause a job to be filtered out.

    # Capture the standard output to verify if the function logs the expected output.
    with io.StringIO() as buf, redirect_stdout(buf):
        # Call the function under test and capture its return value and the standard output.
        result: bool = _should_filter_out_by_tags(job, tags, tags_to_avoid, verbosity)
        run_command_stdout = buf.getvalue().split("\n")

    # Verify the function correctly identifies the job to be filtered out.
    assert result is True

    # Verify the output matches the expected verbosity level.
    if not verbosity:
        assert run_command_stdout == [""]
    else:
        assert run_command_stdout == [
            "runem: not running job 'Job1' because it contains the following tags: "
            "'tag1', 'tag2'",
            "",
        ]


@pytest.mark.parametrize(
    "verbosity",
    [
        True,
        False,
    ],
)
def test_should_filter_out_by_tags_without_tags_to_avoid(verbosity: bool) -> None:
    """Test case where has_tags_to_avoid is empty."""
    job: JobConfig = {
        "label": "Job1",
        "when": {
            "tags": {
                "tag3",
                "tag4",
            }
        },
    }
    tags: JobTags = {"tag3"}
    tags_to_avoid: JobTags = {"tag1", "tag2"}

    with io.StringIO() as buf, redirect_stdout(buf):
        result: bool = _should_filter_out_by_tags(job, tags, tags_to_avoid, verbosity)
        run_command_stdout = buf.getvalue().split("\n")

    if verbosity:
        assert run_command_stdout == [""]
    else:
        assert run_command_stdout == [""]
    assert result is False


@pytest.mark.parametrize(
    "verbosity",
    [
        True,
        False,
    ],
)
@patch("runem.job_filter._should_filter_out_by_tags", return_value=False)
@patch(
    "runem.job_filter.Job.get_job_name", return_value=("intentionally not in job names")
)
def test_get_jobs_matching_when_job_not_in_valid_job_names(
    mock_get_job_name: Mock,
    mock_should_filter: Mock,
    verbosity: bool,
) -> None:
    """Test case where has_tags_to_avoid is not empty."""
    job: JobConfig = {
        "label": "job name not in job_names",
        "when": {"phase": "phase", "tags": set()},
    }
    unused_tags: JobTags = set()
    unused_tags_to_avoid: JobTags = set()
    jobs: PhaseGroupedJobs = defaultdict(list)
    jobs.update({"phase": [job]})

    with io.StringIO() as buf, redirect_stdout(buf):
        _get_jobs_matching(
            phase="phase",
            job_names={"a name that does not match the above job"},
            tags=unused_tags,
            tags_to_avoid=unused_tags_to_avoid,
            jobs=jobs,
            filtered_jobs=jobs,
            verbose=verbosity,
        )
        run_command_stdout = buf.getvalue().split("\n")

    expected_stdout: typing.List[str]
    if not verbosity:
        expected_stdout = [""]
    else:
        expected_stdout = [
            (
                "runem: not running job 'intentionally not in job names' because it "
                "isn't in the list of job names. See --jobs and --not-jobs"
            ),
            "",
        ]
    assert run_command_stdout == expected_stdout
    mock_get_job_name.assert_called_once()
    mock_should_filter.assert_called_once()



"""Job‑typing helpers.

Cross‑version advice
--------------------
* Type variadic keyword arguments as **kwargs: Unpack[KwArgsT] for clarity.
* Always import Unpack from ``typing_extensions``.
  - Std‑lib Unpack appears only in Py 3.12+.
  - ``typing_extensions`` works on 3.9‑3.12, so one import path keeps
    mypy/pyright happy without conditional logic.

Example:
~~~~~~~
from typing_extensions import TypedDict, Unpack


class SaveKwArgs(TypedDict):
    path: str
    overwrite: bool


def save_job(**kwargs: Unpack[SaveKwArgs]) -> None:
    ...
"""

import pathlib
import typing
from datetime import timedelta

from typing_extensions import Unpack

from runem.types.common import FilePathList, PhaseName
from runem.types.options import Options
from runem.types.runem_config import JobConfig

if typing.TYPE_CHECKING:  # pragma: no cover
    from runem.config_metadata import ConfigMetadata

ReportName = str
ReportUrl = typing.Union[str, pathlib.Path]
ReportUrlInfo = typing.Tuple[ReportName, ReportUrl]
ReportUrls = typing.List[ReportUrlInfo]


class JobReturnData(typing.TypedDict, total=False):
    """A dict that defines job result to be reported to the user."""

    reportUrls: ReportUrls  # urls containing reports for the user


TimingEntry = typing.Tuple[str, timedelta]
TimingEntries = typing.List[TimingEntry]


class JobTiming(typing.TypedDict, total=True):
    """A hierarchy of timing info. Job->JobCommands.

    The overall time for a job is in 'job', the child calls to run_command are in
    'commands'
    """

    job: TimingEntry  # the overall time for a job
    commands: TimingEntries  # timing for each call to `run_command`


JobReturn = typing.Optional[JobReturnData]
JobRunMetadata = typing.Tuple[JobTiming, JobReturn]
JobRunTimesByPhase = typing.Dict[PhaseName, typing.List[JobTiming]]
JobRunReportByPhase = typing.Dict[PhaseName, ReportUrls]
JobRunMetadatasByPhase = typing.Dict[PhaseName, typing.List[JobRunMetadata]]


class CommonKwargs(
    typing.TypedDict,
    total=True,  # each of these are guaranteed to exist in jobs and hooks
):
    """Defines the base args that are passed to all jobs.

    As we call hooks and job-task in the same manner, this defines the variables that we
    can access from both hooks and job-tasks.
    """

    config_metadata: "ConfigMetadata"  # gives greater context to jobs and hooks
    job: JobConfig  # the job or hook task spec ¢ TODO: rename this
    label: str  # the name of the hook or the job-label
    options: Options  # options passed in on the command line
    procs: int  # the max number of concurrent procs to run
    root_path: pathlib.Path  # the path where the .runem.yml file is
    verbose: bool  # control log verbosity


class HookSpecificKwargs(typing.TypedDict, total=False):
    """Defines the args that are passed down to the hooks.

    NOTE: that although these however
     outside of the *hook* context, the data will not be present. Such is the
     difficulty in dynamic programming.
    """

    wall_clock_time_saved: timedelta  # only on `HookName.ON_EXIT`


class JobTaskKwargs(
    typing.TypedDict,
    total=False,  # for now, we don't enforce these types for job-context, but we should.
):
    """Defines the task-specific args for job-task functions."""

    file_list: FilePathList
    record_sub_job_time: typing.Optional[typing.Callable[[str, timedelta], None]]


class HookKwargs(CommonKwargs, HookSpecificKwargs):
    """A merged set of kwargs for runem-hooks."""

    pass


class JobKwargs(CommonKwargs, JobTaskKwargs):
    """A merged set of kwargs for job-tasks."""

    pass


class AllKwargs(CommonKwargs, JobTaskKwargs, HookSpecificKwargs):
    """A merged set of kwargs for al job-functions."""

    pass


@typing.runtime_checkable
class JobFunction(typing.Protocol):
    def __call__(self, **kwargs: Unpack[AllKwargs]) -> JobReturn:  # pragma: no cover
        """Defines the call() protocol's abstract pattern for job-tasks."""

    @property
    def __name__(self) -> str:  # pragma: no cover
        """Defines the name protocol for job-task functions.

        This is primarily used for internal tests but can be useful for introspection.
        """


def _hook_example(
    wall_clock_time_saved: timedelta,
    **kwargs: typing.Any,
) -> None:
    """An example hook."""


def _job_task_example(
    **kwargs: Unpack[JobKwargs],
) -> None:
    """An example job-task function."""



import argparse
import pathlib
import typing

from runem.informative_dict import InformativeDict
from runem.types.common import JobNames, JobPhases, JobTags, OrderedPhases
from runem.types.filters import TagFileFilters
from runem.types.options import OptionsWritable
from runem.types.runem_config import OptionConfigs, PhaseGroupedJobs

if typing.TYPE_CHECKING:  # pragma: no cover
    from runem.hook_manager import HookManager


class ConfigMetadata:
    """Full metadata about what can and should be run."""

    phases: OrderedPhases  # the phases and orders to run them in
    options_config: OptionConfigs  # the options to add to the cli and pass to jobs
    file_filters: TagFileFilters  # which files to get for which tag
    hook_manager: "HookManager"  # the hooks register for the run
    jobs: PhaseGroupedJobs  # the jobs to be run ordered by phase
    all_job_names: JobNames  # the set of job-names
    all_job_phases: JobPhases  # the set of job-phases (should be subset of 'phases')
    all_job_tags: JobTags  # the set of job-tags (used for filtering)

    options: OptionsWritable  # the final configured options to pass to jobs

    args: argparse.Namespace  # the raw cli args, probably missing information
    jobs_to_run: JobNames  # superset of job-name candidates to run, from cli+config
    phases_to_run: JobPhases  # superset of phase candidates to run, from cli+config
    tags_to_run: JobTags  # superset of tag-candidates to run, from cli+config
    tags_to_avoid: JobTags  # superset of tag-withhold to NOT run, from cli+config

    def __init__(
        self,
        cfg_filepath: pathlib.Path,
        phases: OrderedPhases,
        options_config: OptionConfigs,
        file_filters: TagFileFilters,
        hook_manager: "HookManager",
        jobs: PhaseGroupedJobs,
        all_job_names: JobNames,
        all_job_phases: JobPhases,
        all_job_tags: JobTags,
    ) -> None:
        self.cfg_filepath = cfg_filepath
        self.phases = phases
        self.options_config = options_config
        self.file_filters = file_filters
        # initialise all the registered call-back hooks
        self.hook_manager = hook_manager
        self.jobs = jobs
        self.all_job_names = all_job_names
        self.all_job_phases = all_job_phases
        self.all_job_tags = all_job_tags

        self.options = InformativeDict()  # shows useful errors on bad-option lookups

        self.args = (
            argparse.Namespace()
        )  # the raw cli args, defined after argument parsing
        self.jobs_to_run = set()  # will be defined after cli arg parsing
        self.phases_to_run = set()  # will be defined after cli arg parsing
        self.tags_to_run = set()  # will be defined after cli arg parsing
        self.tags_to_avoid = set()  # will be defined after cli arg parsing

    def set_cli_data(
        self,
        args: argparse.Namespace,
        jobs_to_run: JobNames,
        phases_to_run: JobPhases,
        tags_to_run: JobTags,
        tags_to_avoid: JobTags,
        options: OptionsWritable,
    ) -> None:
        self.options = options
        self.args = args
        self.jobs_to_run = jobs_to_run
        self.phases_to_run = phases_to_run
        self.tags_to_run = tags_to_run
        self.tags_to_avoid = tags_to_avoid



import pathlib
from datetime import timedelta

from typing_extensions import Unpack

from runem.types import HookKwargs


def _on_exit_hook(
    **kwargs: Unpack[HookKwargs],
) -> None:
    """A noddy hook."""
    assert "wall_clock_time_saved" in kwargs
    wall_clock_time_saved: timedelta = kwargs["wall_clock_time_saved"]
    root_path: pathlib.Path = pathlib.Path(__file__).parent.parent.parent
    assert (root_path / ".runem.yml").exists()
    times_log: pathlib.Path = root_path / ".times.log"
    with times_log.open("a", encoding="utf-8") as file:
        file.write(f"{str(wall_clock_time_saved.total_seconds())}\n")
